#!/bin/bash
#
# Version: 1.0
# Author: Joe Linoff
# Description:
#
# Setup the /opt/site environment as webdev.
#
# Note that you must run as webdev (or some other role account) to
# allow multiple developers to access the setup. Here is how you
# do it:
#
#    $ sudo useradd webdev
#    $ sudo passwd webdev
#    Changing password for user webdev.
#    New password: webdev*pwd
#    Retype new password: webdev*pwd
#    passwd: all authentication tokens updated successfully.
#
#    $ # Add webdev to the sudoers list
#    $ sudo visudo
#
#    $ su - webdev
#    Password:
#    $ cd tmp
#    $ tar jxf /tmp/mksite-1.0.tar.bz2
#    $ ./install.sh
#    $ firefox https://localhost
#
# For development you can run the django debug server:
#
#    $ cd /opt/site/django/website
#    $ python manage.py runserver 8888
#    $ firefox http://localhost:8888
#
# When you run this tool it will:
#
#    1. verify that all of the necessary system packages are installed
#    2. verify that all of the necessary python packages are installed
#    3. install all of the javascript and CSS packages
#    4. install django
#    5. configure gunicorn (with supervisor to control boot time startup)
#    6. configure the nginx server (for HTTP and HTTPS)
#    7. configure postgresql for the django admin users (fixed pg_hba.conf and postgresql.conf)
#    8. configure the django DB interface (for postgresql), mongodb must be handled separately
#    9. collect the django static data (this is a django thing)
#   10. create the dango app and pre-populate it with some stuff
#   11. create the self signed security certificates (to show how it is done)
#
# When it finishes you will still have to do the django admin
# configuration (probably by running python manage.py syncdb).
#
# This configuration assumes that you will postgresql for django admin
# data and MongDB for the site data based on the idea that there are a
# small number of admins and a large amount of data.
#

# ================================================================
# Definitions
# ================================================================
SITEDIR=/opt/site
WEBSITE_NAME=website
WEBUSER=webdev
WEBUSER_PASSWD='webdev*pwd'

# Don't change anything below this line.
WEBAPP_NAME=webapp
VENVDIR=${SITEDIR}/venv
CONFDIR=${SITEDIR}/conf
DJANGODIR=${SITEDIR}/django
DJANGO_WEBSITE_DIR=${DJANGODIR}/${WEBSITE_NAME}
DJANGO_SETTINGS_DIR=${DJANGO_WEBSITE_DIR}/${WEBSITE_NAME}

SSL_CRT_FILE=${CONFDIR}/${WEBSITE_NAME}.crt
SSL_KEY_FILE=${CONFDIR}/${WEBSITE_NAME}.key

WWWDIR=${SITEDIR}/www
JSDIR=$WWWDIR/js
PKGDIR=$WWWDIR/pkg
CSSDIR=$WWWDIR/css
IMGDIR=$WWWDIR/img

YUMPKGS=(nginx \
    python-pip python-virtualenv python-paramiko \
    python-memcached python-setproctitle \
    pcre libevent memcached \
    openssl \
    gcc \
    supervisor \
    mongodb \
    postgresql postgresql-devel postgresql-server pgadmin3 python-psycopg2 \
    )
PYPKGS=(gunicorn django yolk south flask supervisor pymongo psycopg2 \
    fabric pexpect python-memcached \
    )

# ================================================================
#
# Local functions.
#
# ===============================================================
# Reports whether a yum package is installed.
# Returns true (1) if a package is installed or false (0) otherwise.
function yum-pkg-installed() {
    local pkg="$1"
    rpm -q $pkg >/dev/null 2>&1
    local st=$?
    if (( $st == 0 )) ; then
        echo "1"  # true, installed
    else
        echo "0"
    fi
}

# Report whether a pip package is installed.
# Returns true (1) if it is installed or false (0) otherwise.
function pip-pkg-installed() {
    local pkg="$1"
    pip list | grep -i ^$pkg >/dev/null 2>&1
    local st=$?
    if (( $st == 0 )) ; then
        echo "1"  # true, installed
    else
        echo "0"
    fi
}

# Copy a file if it has changed.
# It will make a backup of the original.
function copy() {
    local src="$1"
    local dst="$2"

    local copy_flag=0
    if [ -f $dst ] ; then
        if ! diff -q $src $dst >/dev/null ; then
            # The files changed, backup the current
            # dst if it exists.
            local num=1
            while (( 1 )) ; do
                local dstx="$dst.orig.$num"
                if [ ! -f $dstx ] ; then
                    cp -v $dst $dstx
                    break
                fi
                num=$(( $num + 1))
            done
            copy_flag=1
        fi
    else
        # dst doesn't exist, always copy
        copy_flag=1
    fi

    if (( $copy_flag )) ; then
        # Copy over the new source.
        echo "$src --> $dst"
        sed -e "s/@@WEBSITE@@/$WEBSITE_NAME/g" \
            -e "s/@@WEBUSER@@/$WEBUSER/g" \
            -e "s/@@WEBUSER_PASSWORD@@/$WEBUSER_PASSWD/g" \
            $src >$dst
    fi
}

# Copy on change.
function sudo-copy() {
    local src="$1"
    local dst="$2"
    local copy_flag=0
    if sudo test -f $dst ; then
        if ! sudo diff -q $src $dst >/dev/null ; then
            # The files changed, backup the current
            # dst if it exists.
            local num=1
            while (( 1 )) ; do
                local dstx="$dst.orig.$num"
                if [ ! -f $dstx ] ; then
                    sudo cp -v $dst $dstx
                    break
                fi
                num=$(( $num + 1))
            done
            copyflag=1
        fi
    else
        # dst doesn't exist, always copy
        copy_flag=1
    fi

    if (( $copy_flag )) ; then
        # Copy over the new source.
        TMPFILE=/tmp/install-$$
        rm --f $TMPFILE
        sed -e "s/@@WEBSITE@@/$WEBSITE_NAME/g" \
            -e "s/@@WEBUSER@@/$WEBUSER/g" \
            -e "s/@@WEBUSER_PASSWORD@@/$WEBUSER_PASSWD/g" \
            $src >$TMPFILE
        sudo cp -v $TMPFILE $dst
        rm -f $TMPFILE
    fi
}

# Move
function sudo-move() {
    local src="$1"
    local dst="$2"
    if [ -f $src ] ; then
        sudo mv -v $src $dst
    fi
}

# Link
function sudo-link() {
    local src="$1"
    local dst="$2"
    if [ ! -f $dst ] ; then
        sudo ln -v -s $src $dst
    fi
}

# Mkdir
function sudo-mkdir() {
    local dir="$1"
    if [ ! -d $dir ] ; then
        sudo mkdir -v -p $dir
    fi
}

# ================================================================
#
# Install the packages and tools.
#
# ================================================================
# Who am i?
echo "STEP: Who am i? $(id)"

me=$(whoami)
if [[ "$me" != "${WEBUSER}" ]] ; then
    echo "ERROR: $0 can only be run by user '${WEBUSER}'."
    exit 1
fi

if ! sudo whoami ; then
    echo "ERROR: $(whoami) is not in the sudoers list."
    echo "       Cannot continue."
    exit 1
fi

# Create the initial site directory.
if [ ! -d ${SITEDIR} ] ; then
    echo "STEP: Creating ${SITEDIR}"
    sudo-mkdir ${SITEDIR}
    sudo chown -R ${WEBUSER}:${WEBUSER} ${SITEDIR}
else
    echo "STEP: Already created ${SITEDIR}"
fi

if [ ! -d ${SITEDIR} ] ; then
    echo "ERROR: directory does not exist: $SITEDIR"
    echo "       This may be because ${WEBUSER} is not in the sudoers list."
    exit 1
fi

# Need wget for subsequent steps.
installed=$(yum-pkg-installed wget)
if (( ! $installed )) ; then
    sudo yum install -y wget
fi

# Install the EPEL repository
if [ ! -f /etc/yum.repos.d/epel.repo ] ; then
    echo "STEP: Installing the EPEL repository"
    sudo wget http://download.fedoraproject.org/pub/epel/6/x86_64/epel-release-6-8.noarch.rpm
    sudo rpm -ivh epel-release-6-8.noarch.rpm
else
    echo "STEP: Already installed the EPEL repository"
fi
yum repolist

# Install other useful tools.
echo "STEP: Installing system packages"
for PKG in ${YUMPKGS[@]} ; do
    printf "%-20s " $PKG
    installed=$(yum-pkg-installed $PKG)
    if (( ! $installed )) ; then
        echo "STEP: Installing system package: $PKG"
        sudo yum install -y $PKG
    else
        echo "STEP: Already installed system package: $PKG"
    fi
done

# Setup python venv
if [ ! -d $VENVDIR/bin ] ; then
    echo "STEP: Setting up the virtual environment in ${SITEDIR}/venv"
    virtualenv ${VENVDIR}
else
    echo "STEP: Already Set up the virtual environment in ${SITEDIR}/venv"
fi

. ${VENVDIR}/bin/activate  # Always set the venv.

# Add in local packages if they have not already been added.
echo "STEP: Installing python packages"
for PKG in ${PYPKGS[@]} ; do
    printf "%-20s " $PKG
    installed=$(pip-pkg-installed $PKG)
    if (( ! $installed )) ; then
        echo "STEP: Installing python package: $PKG"
        pip install $PKG
    else
        echo "STEP: Already installed python package: $PKG"
    fi
done

echo "STEP: List python packages"
pip freeze | sort -fu

# ================================================================
#
# Define the www directory structure.
#
# ================================================================
echo "STEP: Creating the basic directory structure"
DIRS=($JSDIR $PKGDIR $CSSDIR $IMGDIR \
    $SITEDIR/run $SITEDIR/conf $SITEDIR/bin $SITEDIR/logs \
    )
for DIR in ${DIRS[@]} ; do
    if [ ! -d $DIR ] ; then
        mkdir -v -p $DIR
    fi
done

# ================================================================
#
# Populate the web javascript directory.
#
# ================================================================
echo "STEP: Installing jquery javascript files"
pushd $JSDIR

# Get the latest jquery.
# <script type="text/javascript" src="js/jquery-2.1.1.js"></script>
if [ ! -f ./jquery-2.1.1.js ] ; then
    wget http://code.jquery.com/jquery-2.1.1.js -O jquery-2.1.1.js
fi

# <script type="text/javascript" src="js/jquery-2.1.1.min.js"></script>
if [ ! -f ./jquery-2.1.1.min.js ] ; then
    wget http://code.jquery.com/jquery-2.1.1.min.js -O jquery-2.1.1.min.js
fi

# The map is not directly referenced.
if [ ! -f ./jquery-2.1.1.min.map ] ; then
    wget http://code.jquery.com/jquery-2.1.1.min.map -O jquery-2.1.1.min.map
fi

popd

# ================================================================
#
# Populate the web package directory.
#
# ================================================================
echo "STEP: Installing web packages"
pushd $PKGDIR

# Get the latest jquery-ui.
# <script type="text/javascript" src="pkg/jquery-ui-1.10.4/js/jquery-ui-1.10.4.min.js"></script>
if [ ! -f ./jquery-ui-1.10.4/index.html ] ; then
    wget http://jqueryui.com/resources/download/jquery-ui-1.10.4.zip
    unzip jquery-ui-1.10.4.zip
    rm -f jquery-ui-1.10.4.zip
fi

# Get the latest jquery-ui-themes.
# This shows how to access the ui-lightness theme.
# <link rel="stylesheet" type="text/css" href="pkg/jquery-ui-themes-1.10.4/themes/ui-lightness/jquery.ui.theme.css">
# <link rel="stylesheet" type="text/css" href="pkg/jquery-ui-themes-1.10.4/themes/ui-lightness/jquery.ui.min.css">
if [ ! -f ./jquery-ui-themes-1.10.4/package.json ] ; then
    wget http://jqueryui.com/resources/download/jquery-ui-themes-1.10.4.zip
    unzip jquery-ui-themes-1.10.4.zip
    rm -f jquery-ui-themes-1.10.4.zip
fi

# Get the latest tooltipster.
# <link rel="stylesheet" type="text/css" href="pkg/tooltipster-master/css/tooltipster.css">
# <link rel="stylesheet" type="text/css" href="pkg/tooltipster-master/css/themes/tooltipster-light.css">
# <script type="text/javascript" src="pkg/tooltipster-master/js/jquery.tooltipster.min.js"></script>
if [ ! -f ./tooltipster-master/index.html ] ; then
    wget https://github.com/iamceege/tooltipster/archive/master.zip -O tooltipster-master.zip
    unzip tooltipster-master.zip
    rm -f tooltipster-master.zip
fi

# Get the latest crypto package.
# <script type="text/javascript" src="pkg/crypto-js-3.1.2/rollups/aes.js"></script>
# <script type="text/javascript" src="pkg/crypto-js-3.1.2/rollups/md5.js"></script>
# <script type="text/javascript" src="pkg/crypto-js-3.1.2/rollups/sha512.js"></script>
# <script type="text/javascript" src="pkg/crypto-js-3.1.2/rollups/tripledes.js"></script>
if [ ! -f ./crypto-js-3.1.2/components/aes.js ] ; then
    if [ ! -d crypto-js-3.1.2 ] ; then
        mkdir crypto-js-3.1.2
    fi
    cd crypto-js-3.1.2
    wget http://crypto-js.googlecode.com/files/CryptoJS%20v3.1.2.zip -O crypto-js-3.1.2.zip
    unzip crypto-js-3.1.2.zip
    rm -f crypto-js-3.1.2.zip
    cd ..
fi

# Get the latest datatables.
# <link rel="stylesheet" type="text/css" href="pkg/DataTables-1.10.0/media/css/jquery.dataTables.min.css">
# <link rel="stylesheet" type="text/css" href="pkg/DataTables-1.10.0/media/css/jquery.dataTables_themeroller.min.css">
# <script type="text/javascript" src="pkg/DataTables-1.10.0/media/js/jquery.dataTables.min.js"></script>
if [ ! -f ./DataTables-1.10.0/package.json ] ; then
    wget http://datatables.net/releases/DataTables-1.10.0.zip
    unzip DataTables-1.10.0.zip
    rm -f DataTables-1.10.0.zip
fi

# Fancy tree (dynamic display trees).
# <link rel="stylesheet" type="text/css" href="pkg/fancytree-2.0.0/dist/skin-themeroller/ui.fancytree.min.css">
# <script type="text/javascript" src="pkg/fancytree-2.0.0/dist/jquery.fancytree.min.js"></script>
if [ ! -f ./fancytree-2.0.0/README.md ] ; then
    wget https://github.com/mar10/fancytree/archive/v2.0.0.zip -O fancytree-2.0.0.zip
    unzip fancytree-2.0.0.zip
    rm -f fancytree-2.0.0.zip
fi

# Flot plotting tool.
# <script type="text/javascript" src="pkg/flot-0.8.3/jquery.flot.min.js"></script>
# <script type="text/javascript" src="pkg/flot-0.8.3/jquery.flot.pie.min.js"></script>
# <script type="text/javascript" src="pkg/flot-0.8.3/jquery.flot.resize.min.js"></script>
# <script type="text/javascript" src="pkg/flot-0.8.3/jquery.flot.time.min.js"></script>
if [ ! -f ./flot-0.8.3/README.md ] ; then
    wget http://www.flotcharts.org/downloads/flot-0.8.3.zip
    unzip flot-0.8.3.zip
    rm -f flot-0.8.3.zip
    mv flot flot-0.8.3
fi

# Twitter bootstrap framework.
# <link rel="stylesheet" type="text/css" href="pkg/bootstrap-3.1.1-dist/css/bootstrap.min.css">
# <link rel="stylesheet" type="text/css" href="pkg/bootstrap-3.1.1-dist/css/bootstrap-theme.min.css">
# <script type="text/javascript" src="pkg/bootstrap-3.1.1-dist/js/bootstrap.min.js"></script>
if [ ! -f ./bootstrap-3.1.1-dist/js/bootstrap.js ] ; then
    wget https://github.com/twbs/bootstrap/releases/download/v3.1.1/bootstrap-3.1.1-dist.zip
    unzip bootstrap-3.1.1-dist.zip
    rm -f bootstrap-3.1.1-dist.zip
fi

popd

# ================================================================
# Create the django initial project directory
# ================================================================
if [ ! -d ${DJANGO_WEBSITE_DIR} ] ; then
    echo "STEP: Installing django project."
    mkdir -p ${DJANGODIR}
    pushd ${DJANGODIR}
    django-admin.py startproject ${WEBSITE_NAME}
    # TEST Django: http://localhost:8080
    #  $ cd ${WEBSITE_NAME}
    #  $ python manage.py runserver 8080
    #
    # TEST gunicorn: http://localhost:890
    #  # Verify that gunicorn can serve the django app.
    #  $ gunicorn ${WEBSITE_NAME}.wsgi:application --bind 127.0.0.1:8090
    popd
else
    echo "STEP: Django project already installed."
fi

# ================================================================
# Create the gunicorn_startup file.
# ================================================================
GUNICORN_STARTUP=${SITEDIR}/bin/gunicorn_startup.sh
if [ ! -f ${GUNICORN_STARTUP} ] ; then
    echo "STEP: Creating ${GUNICORN_STARTUP}"
    SRC=setup/gunicorn_startup.sh
    if [ ! -f $SRC ] ; then
        echo "ERROR: Missing file $SRC"
        echo "       This indicates a setup failure."
        exit 1
    fi
    copy $SRC ${GUNICORN_STARTUP}
    chmod a+x ${GUNICORN_STARTUP}
else
    echo "STEP: Already created ${GUNICORN_STARTUP}"
fi

# ================================================================
# supervisor
# ================================================================
DST=/etc/supervisord.conf
if ! grep gunicorn_startup $DST >/dev/null ; then
    echo "STEP: Updating $DST to include gunicorn_startup"
    SRC=setup/gunicorn_startup.conf
    if [ ! -f $SRC ] ; then
        echo "ERROR: Missing file $SRC"
        echo "       This indicates a setup failure."
        exit 1
    fi

    touch /opt/site/logs/gunicorn_startup.log
    TMPFILE=local.conf
    cp /etc/supervisord.conf ${TMPFILE}

    sed -e "s/@@WEBSITE@@/$WEBSITE_NAME/g" \
        -e "s/@@WEBUSER@@/$WEBUSER/g" \
        -e "s/@@WEBUSER_PASSWORD@@/$WEBUSER_PASSWD/g" \
        $SRC >>${TMPFILE}

    sudo-copy ${TMPFILE} $DST
    rm -f ${TMPFILE}
    sudo service supervisord start
    sudo service supervisord reload
    sudo service supervisord status
    sudo chkconfig supervisord on
else
    echo "STEP: Already updated $DST to include gunicorn_startup"
fi

# ================================================================
# nginx server
# ================================================================
if [ ! -f /etc/nginx/sites-enabled/nginx-${WEBSITE_NAME}.conf ] ; then
    echo "STEP: Configuring nginx"
    mkdir -p $DJANGODIR/static
    mkdir -p $DJANGODIR/media

    sudo-copy setup/nginx.conf /etc/nginx/nginx.conf 
    sudo-move /etc/nginx/conf.d/default.conf{,.orig}
    sudo-mkdir /etc/nginx/sites-available
    sudo-mkdir /etc/nginx/sites-enabled
    sudo-copy setup/nginx-${WEBSITE_NAME}.conf /etc/nginx/sites-available/nginx-${WEBSITE_NAME}.conf
    sudo-link /etc/nginx/sites-available/nginx-${WEBSITE_NAME}.conf /etc/nginx/sites-enabled/nginx-${WEBSITE_NAME}.conf

    sudo service nginx restart
    sudo chkconfig nginx on
else
    echo "STEP: Already configured nginx"
fi

# ================================================================
# Postgres setup for django admins.
# MongoDB is used for the actual data.
# ================================================================
if ! sudo test -f /var/lib/pgsql/data/pg_hba.conf ; then
    echo "STEP: Initialize database"
    sudo service postgresql initdb
    sudo-copy setup/postgres-pg_hba.conf /var/lib/pgsql/data/pg_hba.conf
    sudo-copy setup/postgres-postgresql.conf /var/lib/pgsql/data/postgresql.conf
    sudo service postgresql restart

    # The website database is created by the postgres user so
    # it must be created in a directory where postgres can read
    # and write.
    copy setup/postgres-setup.sql /tmp/x.sql
    chmod a+rw /tmp/x.sql
    pushd /tmp
    sudo -u postgres psql -f /tmp/x.sql
    rm -f /tmp/x.sql
    popd

    # Later this will be synced with Django.
else
    echo "STEP: Already initialized database"
fi

# ================================================================
# Django DB setup.
# ================================================================
if [ ! -f ${DJANGO_SETTINGS_DIR}/.mksite-dbsetup ] ; then
    echo "STEP: Syncing up the Django database"
    copy setup/django-settings-initial.py ${DJANGO_SETTINGS_DIR}/settings.py
    pushd $DJANGO_WEBSITE_DIR
    python manage.py syncdb --noinput
    popd
    touch ${DJANGO_SETTINGS_DIR}/.mksite-dbsetup
else
    echo "STEP: Already synced up the Django database"
fi

# ================================================================
# Collect static data.
# ================================================================
echo "STEP: Collect static data"
pushd $DJANGO_WEBSITE_DIR
python manage.py collectstatic --noinput
popd

# ================================================================
# Create the app.
# ================================================================
if [ ! -d $DJANGO_WEBSITE_DIR/$WEBAPP_NAME ] ; then
    echo "STEP: Create the app: $WEBAPP_NAME"
    copy setup/django-settings-final.py ${DJANGO_SETTINGS_DIR}/settings.py
    copy setup/django-urls.py ${DJANGO_SETTINGS_DIR}/urls.py
    cp setup/django-webapp.tar.bz2 $DJANGO_WEBSITE_DIR/
    pushd $DJANGO_WEBSITE_DIR
    python manage.py startapp $WEBAPP_NAME
    tar jxf django-webapp.tar.bz2
    rm -f django-webapp.tar.bz2
    popd
else
    echo "STEP: Already created the app: $WEBAPP_NAME"
fi

# ================================================================
# Create the signed security certificates.
# ================================================================
if [ ! -f $SSL_CRT_FILE ] ; then
    echo "STEP: Creating the SSL certificate: $SSL_CRT_FILE"
    openssl req -new -newkey rsa:4096 -days 365 -nodes -x509 \
        -subj "/C=US/ST=Washington/L=Seattle/O=${WEBSITE_NAME}/OU=${WEBAPP_NAME}/CN=localhost" \
        -keyout $SSL_KEY_FILE \
        -out $SSL_CRT_FILE
else
    echo "STEP: Already created the SSL certificate: $SSL_CRT_FILE"
fi

# ================================================================
# Setup the services for re-entrant runs.
# ================================================================
echo "STEP: Restarting system services for re-entrant runs"
sudo service postgresql restart
sudo service supervisord restart
sudo service nginx restart

# ================================================================
# Done.
# ================================================================
echo "STEP: Done"
cat <<EOF

The installation is complete. You should now be able to run
http://localhost or https://localhost to see the newly installed
website.

EOF
